"use strict";
/* IMPORT */
const promise_make_naked_1 = require("promise-make-naked");
const worker_1 = require("./worker");
/* WORKTANK */
class WorkTank {
    /* CONSTRUCTOR */
    constructor(options) {
        var _a, _b;
        this.terminated = false;
        this.name = (_a = options.name) !== null && _a !== void 0 ? _a : 'WorkTank-Worker';
        this.size = (_b = options.size) !== null && _b !== void 0 ? _b : 1;
        this.methods = this._getMethodsSerialized(options.methods);
        this.tasksBusy = new Set();
        this.tasksReady = new Set();
        this.workersBusy = new Set();
        this.workersReady = new Set();
    }
    /* HELPERS */
    _getMethodsSerialized(methods) {
        if (typeof methods === 'string') { // Serialized function that returns the methods, useful for complex workers
            return methods;
        }
        else { // Deserialized methods map
            const serialized = {};
            for (const method in methods) {
                serialized[method] = methods[method].toString();
            }
            return serialized;
        }
    }
    _getTaskReady() {
        for (const task of this.tasksReady)
            return task;
    }
    _getWorkerName() {
        if (this.size < 2)
            return this.name;
        const counter = 1 + (this.workersBusy.size + this.workersReady.size);
        return `${this.name} (${counter})`;
    }
    _getWorkerReady() {
        for (const worker of this.workersReady)
            return worker;
        if (this.workersBusy.size >= this.size)
            return;
        const name = this._getWorkerName();
        const worker = new worker_1.default(this.methods, name);
        this.workersReady.add(worker);
        return worker;
    }
    /* API */
    exec(method, args) {
        const { promise, resolve, reject } = promise_make_naked_1.default(), task = { method, args, promise, resolve, reject };
        this.terminated = false;
        this.tasksReady.add(task);
        this.tick();
        return promise;
    }
    terminate() {
        this.terminated = true;
        /* RESETTING TASKS */
        const error = new Error('WorkTank terminated!');
        for (const task of this.tasksBusy)
            task.reject(error);
        for (const task of this.tasksReady)
            task.reject(error);
        this.tasksBusy = new Set();
        this.tasksReady = new Set();
        /* RESETTING WORKERS */
        for (const worker of this.workersBusy)
            worker.terminate();
        for (const worker of this.workersReady)
            worker.terminate();
        this.workersBusy = new Set();
        this.workersReady = new Set();
    }
    tick() {
        const worker = this._getWorkerReady();
        if (!worker)
            return;
        const task = this._getTaskReady();
        if (!task)
            return;
        this.workersReady.delete(worker);
        this.workersBusy.add(worker);
        this.tasksReady.delete(task);
        this.tasksBusy.add(task);
        task.promise.finally(() => {
            if (this.terminated)
                return;
            this.workersBusy.delete(worker);
            this.workersReady.add(worker);
            this.tasksBusy.delete(task);
            this.tick();
        });
        worker.exec(task);
    }
}
;
/* EXPORT */
module.exports = WorkTank;
module.exports.default = WorkTank;
Object.defineProperty(module.exports, "__esModule", { value: true });
